/* dutree  -- Graphical display of disk useage
   Copyright (C) 1992, 1996, 2000 Free Software Foundation, Inc.
   Copyright (C) 2002 Aaron Bentley

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

#include <inttypes.h>
#define xmalloc malloc

/**
 * \brief Represents a directory entry
 */
class EnTree
{
public:
   ///The name of the directory
   char *name;
   ///The size of the directory (and all children)
   uintmax_t size;
   ///The parent of the directory
   EnTree *parent;
   ///The first child of the directory
   EnTree *child;
   ///The next sibling of this directory
   EnTree *next;
   ///The previous sibling of this direcotry
   EnTree *prev;
   int r, g, b;
   EnTree()
   {
      init();
   }
/**
 * \brief Initialize a node to safe defaults 
 */
   void init()
   {
      EnTree *target=this;
      if (target==NULL) return;
      target->size=0;
      target->next=NULL;
      target->prev=NULL;
      target->parent=NULL;
      target->child=NULL;
      target->name=NULL;
      target->r=random()%65535;
      target->g=random()%65535;
      target->b=random()%65535;
   }
   /**
    * \brief Destroys this node and all its descendants
    */
   ~EnTree()
   {
      if (this==NULL) return;
      EnTree *cur;
      for (cur=child; cur!=NULL; cur=cur->next)
      {
	 delete cur;
      }
      free(name);
   }

/**
 * \brief Find the end of the list
 */
   EnTree *seek_end();
   
   /* Add a child to a node */
   int add_child(EnTree *child);

   void remove();

   void set(uintmax_t n_blocks, const char *name);

   void add(uintmax_t n_blocks, const char *name);

   void incr_up(uintmax_t n_blocks);
};

/**
 * \brief Find the last sibling of this EnTree
 * \return The last EnTree
 */
EnTree *EnTree::seek_end()
{
   EnTree *current;
   for (current=this; current->next!=NULL; current=current->next);
   return current;
}

/**
 * \brief Add a child to this node
 * \param child The child to add
 */
int EnTree::add_child(EnTree *child)
{ 
   if (this->child==NULL) this->child=child;
   else
   {
      EnTree *end;
      end=this->child->seek_end();
      end->next=child;
      child->prev=end;
   }
   child->parent=this;
   return 0;
}

/**
 * \brief Removes this EnTree from its tree without deleting it
 */
void EnTree::remove()
{
   // remove reference in the previous node
   if (prev!=NULL) prev->next=next;

   // remove reference in the next node
   if (next!=NULL) next->prev=prev;

   // remove reference in the parent node
   if (parent!=NULL && parent->child==this) 
   parent->child=next;
}

/**
 * \brief Set the name and size of an EnTree node 
 * \param n_blocks The size of the entry
 * \param name The name of the entry
 */
void EnTree::set(uintmax_t n_blocks, const char *name)
{
   size=n_blocks;
   if (this->name!=NULL) free (this->name);
   if (name==NULL) this->name=NULL;
   else this->name=strdup(name);
}

/**
 * \brief Set the name and increments the size of an EnTree node
 * \param n_blocks The size to add
 * \param name The name to set
 */
void EnTree::add(uintmax_t n_blocks, const char *name)
{
   this->size+=n_blocks;
   if (this->name!=NULL) free (this->name);
   if (name==NULL) this->name=NULL;
   else this->name=strdup(name);
}

/**
 * \brief increases the size of all ancestors of this block
 * \param n_blocks The size to increase by
 */
void EnTree::incr_up(uintmax_t n_blocks)
{
  for (EnTree *cur=parent; cur!=NULL; cur=cur->parent)
  {
	  cur->size+=n_blocks;
  }
}
